﻿Option Explicit On
Option Strict On

Imports System.Data
Imports System.Data.Sql
Imports System.Data.SqlClient
Imports Microsoft.VisualBasic

''' <summary>
''' Allows creating, updating or deleting a (proprietary) lock on a single record.
''' Locks are for SQL Server only and will be kept in the table tblRecord_Locks.
''' </summary>
''' <remarks></remarks>
Public Class cSQLS_LockHandler
   Inherits cSQLS_Base
   Implements IDisposable

#Region " --- Member Variables --- "

   ''' <summary>
   ''' The name of the table that is being used with this class. To be passed in the constructor.
   ''' </summary>
   ''' <remarks></remarks>
   Private _strTableName As String

   'All published via properties (see there)
   Private _intLockSettings_Duration_Seconds As Integer = 600
   Private _intLockSettings_ShowMessageForm_BeforeLockExpires_Seconds As Integer = 60
   Private _intLockSettings_SP_AutoUnlock_Threshold_Seconds As Integer = 30
   Private _frmCaller As Form = Nothing
   Private _frmLockWatcher As frm03b_Lock_ExpirationWarning = Nothing

   ''' <summary>
   ''' Signal to callers: True, if the lock was removed automatically, false if not (i.e. the user decided to remove the lock)
   ''' </summary>
   ''' <remarks></remarks>
   Private _fLockAutoRemoved As Boolean = False

#End Region

#Region " --- Events raised by this class --- "

   Public Event LockRemoved(ByVal TableName As String, ByVal RecordID As Integer, ByVal LockAutoRemoved As Boolean)

#End Region

#Region " --- Properties --- "

   Public Property LockAutoRemoved() As Boolean
      Get
         Return _fLockAutoRemoved
      End Get
      Set(ByVal value As Boolean)
         _fLockAutoRemoved = value
      End Set
   End Property

   ''' <summary>
   ''' The time a lock is to be kept up before informing the user of its
   ''' upcoming expiration.
   ''' </summary>
   Public ReadOnly Property LockSettings_Duration_Seconds() As Integer
      Get
         Return _intLockSettings_Duration_Seconds
      End Get
   End Property

   ''' <summary>
   ''' The number of seconds before users are notified of an expiring lock.
   ''' </summary>
   Public ReadOnly Property LockSettings_ShowMessageForm_BeforeLockExpires_Seconds() As Integer
      Get
         Return _intLockSettings_ShowMessageForm_BeforeLockExpires_Seconds
      End Get
   End Property

   ''' <summary>
   ''' Use deprecated!
   ''' </summary>
   ''' <value></value>
   Public ReadOnly Property LockSettings_SP_AutoUnlock_Threshold_Seconds() As Integer
      Get
         Return _intLockSettings_SP_AutoUnlock_Threshold_Seconds
      End Get
   End Property

   ''' <summary>
   ''' The form that is to be shown when users are to be informed of an
   ''' expiring lock.
   ''' </summary>
   Public Property LockWatcherForm() As frm03b_Lock_ExpirationWarning
      Get
         Return _frmLockWatcher
      End Get
      Set(ByVal value As frm03b_Lock_ExpirationWarning)
         _frmLockWatcher = value
      End Set
   End Property

   ''' <summary>
   ''' The form that owns a lock.
   ''' </summary>
   Public Property CallerForm() As Form
      Get
         Return _frmCaller
      End Get
      Set(ByVal value As Form)
         _frmCaller = value
      End Set
   End Property

   Public Property TableName() As String
      Get
         Return _strTableName
      End Get
      Set(ByVal value As String)
         _strTableName = value
      End Set
   End Property

#End Region

#Region " --- Methods --- "

   ''' <summary>
   ''' Attempts to create a lock for the given record. 
   ''' The lock will be created for the current user's Windows logon-name and with the system date/time as current on the SQL Server.
   ''' If a lock already exists, its timestamp will be updated.
   ''' </summary>
   ''' <param name="intRecord_PK">The ID of the record that is to be locked.</param>
   ''' <returns>True: Lock was created successfully; 
   ''' False: Lock could not be created (i.e. a different user already locked this record) - in this case, a message will have been shown
   ''' with information about the user that originally locked the record, including the date/time when the lock was created.</returns>
   ''' <remarks></remarks>
   Function Create_Lock(ByVal intRecord_PK As Integer) As Boolean
      Dim fConnectionOpenedHere_CloseWhenExiting As Boolean = (DB_Connection.State <> ConnectionState.Open)
      Try
         Dim dt As New DataTable(Me.ToString & "_Lock")
         Dim cmd As New SqlCommand("dbo.p_Lock_Attempt_Create", DB_Connection)
         cmd.CommandType = CommandType.StoredProcedure
         cmd.Parameters.Add("@nvcTableName", SqlDbType.NVarChar)
         cmd.Parameters.Add("@nvcUser_WinLogon", SqlDbType.NVarChar)
         cmd.Parameters.Add("@intRecord_PK", SqlDbType.Int)

         cmd.Parameters("@nvcTableName").Value = TableName
         cmd.Parameters("@nvcUser_WinLogon").Value = cUserProps.Windows_LogonName()
         cmd.Parameters("@intRecord_PK").Value = intRecord_PK

         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Open Then DB_Connection.Open()

         Dim da As New SqlDataAdapter(cmd)
         da.Fill(dt)
         DB_Connection.Close()

         'The SP will return a table in which the neuralgic information is to be taken from the field "Lock_Created" - this field
         'will indicate whether or not a lock was created.
         If CType(dt.Rows(0)("Lock_Created"), Boolean) Then
            'OK - lock was created.

            'If, when calling this class'es constructor, a calling form has been defined, activate the auto-unlocker

            'Read the current lock-settings and store them locally
            If Not Me.LockWatcherForm Is Nothing Then
               Me.LockWatcherForm.CloseFormWithoutAction()
               Application.DoEvents()
            End If
            LockSettings_Read()
            _frmLockWatcher = New frm03b_Lock_ExpirationWarning(Me, intRecord_PK)

            Return True
         Else
            'Lock already existing - inform user and exit.
            MsgBox("This record has already been locked by user """ & dt.Rows(0)("Lock_User").ToString & """." & vbCrLf & _
               "The lock was created on " & CType(dt.Rows(0)("Lock_Datetime"), Date).ToString("dd.MM.yyyy\, HH:mm:ss") & "." & vbCrLf & vbCrLf & _
               "Please try again later. Should the lock persist, please inform your system-administrator.", _
               MsgBoxStyle.Exclamation, Application.ProductName & ": Record already locked!")
            Return False
         End If

      Catch ex As Exception
         MsgBox("The following error occured while attempting to lock the record for editing:" & vbCrLf & vbCrLf & _
            ex.Message, MsgBoxStyle.Exclamation, Application.ProductName & ": Error")
         Return False
      Finally
         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Closed Then DB_Connection.Close()
      End Try
   End Function

   ''' <summary>
   ''' Attempts to remove the lock for the given record (using the instance's TableName property and the GUID of the record that is locked in the respective table). 
   ''' </summary>
   ''' <param name="intRecord_PK">The ID of the record in the target table that is to be unlocked.</param>
   ''' <param name="fShowMessageIfFailed">Allows preventing a failure-message from popping up in the case that the lock could not be removed for whatever reason.</param>
   ''' <returns>True: Lock was removed successfully; 
   ''' False: Lock could not be removed (i.e. due to the lock no longer existing or the fact that the current user is not the originator of the lock and
   ''' does not have sufficient privileges to remove the lock, or due to a server-error).
   ''' If the lock could not be removed, a message will have been shown.</returns>
   ''' <remarks></remarks>
   Function Remove_Lock(ByVal intRecord_PK As Integer, Optional ByVal fShowMessageIfFailed As Boolean = True) As Boolean
      Static sfInHere As Boolean

      If sfInHere Then Return True
      sfInHere = True
      Dim fConnectionOpenedHere_CloseWhenExiting As Boolean = (DB_Connection.State <> ConnectionState.Open)
      Dim dr As SqlDataReader = Nothing
      Try
         Dim cmd As New SqlCommand("dbo.p_Lock_Attempt_Remove", DB_Connection)
         cmd.CommandType = CommandType.StoredProcedure
         cmd.Parameters.Add("@nvcTableName", SqlDbType.NVarChar)
         cmd.Parameters.Add("@nvcUser_WinLogon", SqlDbType.NVarChar)
         cmd.Parameters.Add("@intRecord_PK", SqlDbType.Int)

         cmd.Parameters("@nvcTableName").Value = TableName
         cmd.Parameters("@nvcUser_WinLogon").Value = cUserProps.Windows_LogonName()
         cmd.Parameters("@intRecord_PK").Value = intRecord_PK

         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Open Then DB_Connection.Open()

         dr = cmd.ExecuteReader
         If dr.HasRows Then
            dr.Read()
            If CType(dr("Lock_StillExists"), Boolean) = False Then
               'OK - lock was removed (or at least doesn't exist - simply discard information if i.e. the lock didn't exist at all).

               If Not Me.LockWatcherForm Is Nothing Then Me.LockWatcherForm.CloseFormWithoutAction()
               RaiseEvent LockRemoved(Me.TableName, intRecord_PK, False)
               Return True
            Else
               'Lock still existing - inform user about the reason (as returned from the SP) and exit.
               MsgBox("The record could not be unlocked due to the following reason:" & vbCrLf & _
                  dr("Lock_NotRemoved_Reason").ToString & vbCrLf & vbCrLf & _
                  "Please inform your system-administrator if you think you should have been able to remove the lock.", _
                  MsgBoxStyle.Exclamation, Application.ProductName & ": Record already locked!")
               Return False
            End If
         Else
            Return False
         End If
         DB_Connection.Close()

      Catch ex As Exception
         MsgBox("The following error occured while attempting to unlock a record:" & vbCrLf & vbCrLf & _
            ex.Message, MsgBoxStyle.Exclamation, Application.ProductName & ": Error")
         Return False
      Finally
         If Not dr Is Nothing AndAlso Not dr.IsClosed Then dr.Close()
         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Closed Then DB_Connection.Close()
         sfInHere = False
      End Try
   End Function

   ''' <summary>
   ''' Attempts to remove the lock for the given record (using the Lock Record's GUID itself). 
   ''' </summary>
   ''' <param name="Lock_GUID">The GUID of the record in the Lock Records table that is to be unlocked.</param>
   ''' <param name="fShowMessageIfFailed">Allows preventing a failure-message from popping up in the case that the lock could not be removed for whatever reason.</param>
   ''' <returns>True: Lock was removed successfully; 
   ''' False: Lock could not be removed (i.e. due to the lock no longer existing or the fact that the current user is not the originator of the lock and
   ''' does not have sufficient privileges to remove the lock, or due to a server-error).
   ''' If the lock could not be removed, a message will have been shown.</returns>
   ''' <remarks></remarks>
   Function Remove_Lock_ByLockGUID(ByVal Lock_GUID As Guid, Optional ByVal fShowMessageIfFailed As Boolean = True) As Boolean
      Dim fConnectionOpenedHere_CloseWhenExiting As Boolean = (DB_Connection.State <> ConnectionState.Open)
      Dim dr As SqlDataReader = Nothing
      Try
         Dim cmd As New SqlCommand("dbo.p_Lock_Attempt_Remove_ByLockGuid", DB_Connection)
         cmd.CommandType = CommandType.StoredProcedure
         cmd.Parameters.Add("@Lock_GUID", SqlDbType.UniqueIdentifier)
         cmd.Parameters("@Lock_GUID").Value = Lock_GUID

         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Open Then DB_Connection.Open()

         dr = cmd.ExecuteReader
         If dr.HasRows Then
            dr.Read()

            If CType(dr("Locks_Remaining"), Integer) = 0 Then
               If Not Me.LockWatcherForm Is Nothing Then Me.LockWatcherForm.CloseFormWithoutAction()
               RaiseEvent LockRemoved(dr("Lock_Table").ToString, CType(dr("Lock_Record_PK"), Integer), False)
               Return True
            Else
               Return False
            End If
         Else
            Return False
         End If

      Catch ex As Exception
         MsgBox("The following error occured while attempting to remove a Record Lock:" & vbCrLf & vbCrLf & _
            ex.Message, MsgBoxStyle.Exclamation, Application.ProductName & ": Error")
         Return False
      Finally
         If Not dr Is Nothing AndAlso Not dr.IsClosed Then dr.Close()
         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Closed Then DB_Connection.Close()
      End Try
   End Function

   '*** DB-specific list of current locks: Commented out for this sample! ***
   '''' <summary>
   '''' Returns a DataTable with all Record Locks, including formatted information that may be presented to users
   '''' </summary>
   'Public Function Records_DataTable() As DataTable
   '   Dim dt As New DataTable(Me.ToString)
   '   Dim fConnectionOpenedHere_CloseWhenExiting As Boolean = (DB_Connection.State <> ConnectionState.Open)
   '   Try
   '      Dim cmd As New SqlCommand("dbo.p_Locks_RecordList", DB_Connection)
   '      cmd.CommandType = CommandType.StoredProcedure

   '      If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Open Then DB_Connection.Open()

   '      Dim da As New SqlDataAdapter(cmd)
   '      da.Fill(dt)
   '      Return dt

   '   Catch ex As Exception
   '      MsgBox("The following error occured while attempting to unlock a record:" & vbCrLf & vbCrLf & _
   '         ex.Message, MsgBoxStyle.Exclamation, Application.ProductName & ": Error")
   '      Return dt
   '   Finally
   '      If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Closed Then DB_Connection.Close()
   '   End Try
   'End Function

   ''' <summary>
   ''' Reads the current lock-settings and stores them locally
   ''' </summary>
   ''' <remarks></remarks>
   Private Sub LockSettings_Read()
      Dim fConnectionOpenedHere_CloseWhenExiting As Boolean = (DB_Connection.State <> ConnectionState.Open)
      Dim dr As SqlDataReader = Nothing
      Try
         Dim fSuccess As Boolean = False
         Dim cmd As New SqlCommand("SELECT TOP 1 * FROM dbo.tblRecord_Locks_Settings", DB_Connection)
         cmd.CommandType = CommandType.Text

         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Open Then DB_Connection.Open()
         dr = cmd.ExecuteReader
         If dr.HasRows Then
            dr.Read()
            _intLockSettings_Duration_Seconds = CType(dr("LS_MaxLockTime_Seconds"), Integer)
            _intLockSettings_ShowMessageForm_BeforeLockExpires_Seconds = CType(dr("LS_ShowForm_SecondsBeforeExpiring"), Integer)
            _intLockSettings_SP_AutoUnlock_Threshold_Seconds = CType(dr("LS_SP_Unlock_Threshold"), Integer)
         End If

      Catch ex As Exception
         MsgBox("Fehler beim Lesen der Einstellungen für Datensatzsperren:" & vbCrLf & vbCrLf & _
            ex.Message, MsgBoxStyle.Exclamation, Application.ProductName & ": Fehler")
         Return
      Finally
         If Not dr Is Nothing AndAlso Not dr.IsClosed Then dr.Close()
         If fConnectionOpenedHere_CloseWhenExiting AndAlso DB_Connection.State <> ConnectionState.Closed Then DB_Connection.Close()
      End Try
   End Sub

#End Region

#Region " --- Constructors --- "

   ''' <summary>
   ''' Constructor.
   ''' </summary>
   ''' <param name="strTableName">The name of the table that locks are to be administered for.</param>
   ''' <remarks></remarks>
   Public Sub New(ByVal strTableName As String)
      TableName = strTableName
   End Sub

   ''' <summary>
   ''' Constructor.
   ''' </summary>
   ''' <param name="strTableName">The name of the table that locks are to be administered for.</param>
   ''' <remarks></remarks>
   Public Sub New(ByVal strTableName As String, ByRef frmCaller As Form)
      Me.New(strTableName)
      CallerForm = frmCaller
   End Sub

#End Region

#Region " --- IDisposable (implementation) --- "

   Private _fdisposedValue As Boolean = False      ' To detect redundant calls
   Protected Overridable Sub Dispose(ByVal disposing As Boolean)
      If Not _fdisposedValue Then
         If disposing Then
            If Not Me.LockWatcherForm Is Nothing Then
               Try
                  Me.LockWatcherForm.CloseFormWithoutAction()
               Catch ex As Exception
               End Try
            End If
            ' TODO: free other state (managed objects).
         End If

         ' TODO: free your own state (unmanaged objects).
         ' TODO: set large fields to null.
      End If
      _fdisposedValue = True
   End Sub

   Public Sub Dispose() Implements IDisposable.Dispose
      ' This code added by Visual Basic to correctly implement the disposable pattern.
      ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
      Dispose(True)
      GC.SuppressFinalize(Me)
   End Sub

#End Region

End Class
