Option Explicit On
Option Strict On

''' <summary>
''' Provides a generic approach to data-entry forms that consist of a DataGridView control
''' that contains all records plus detail-controls that allow editing once an item in the
''' grid has been selected by the user.
''' </summary>
Public Class frm03a_BASE_DataForm

#Region " --- Must Override --- "

   '**** ATTENTION ****
   '
   'Usually, all the following properties and methods would be declared as <MustOverride> (and this class would carry the <MustInherit>-attribute, 
   'but due to a problem in the .net 2.0/3.0/3.5 framework, the Windows Forms Designer would then not be able to display the form.
   'The only work-around I know of exists for C# (the <Abstract>-attribute is available there), but there doesn't seem to be a way to do this
   'in VB.net.
   'As a result, the methods and properties in this region are rather declared as <Overridable> and need to be carrying the <Overrides>-attribute
   'in the inheriting form. Methods and properties that require overriding will throw an exception (at runtime) if they haven't actually been overrided
   'by the inheritor.

   ''' <summary>
   ''' This property is to return the designator of a record in singular, i.e. "user" or "conversation item"
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable ReadOnly Property RecordDesignator_Singular() As String
      Get
         Return "record"
      End Get
   End Property

   ''' <summary>
   ''' This property is to return the designator of a record in plural, i.e. "users" or "conversation items"
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable ReadOnly Property RecordDesignator_Plural() As String
      Get
         Return "records"
      End Get
   End Property

   ''' <summary>
   ''' This property must return the index of the column within the records-grid that contains the record's primary key (String).
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable ReadOnly Property GridControl() As DataGridView
      Get
         Throw New NotImplementedException("GridControl-Property has not been provided by the inheritor.")
         Return Nothing
      End Get
   End Property

   ''' <summary>
   ''' This property must return the index of the column within the records-grid that contains the record's primary key (String).
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable ReadOnly Property Grid_ColumnIndex_Record_ID() As Integer
      Get
         Throw New NotImplementedException("Grid_ColumnIndex_Record_ID-Property has not been provided by the inheritor.")
         Return 0
      End Get
   End Property

   ''' <summary>
   ''' Must return a reference to the LockHandler
   ''' </summary>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable Property LockHandler() As cSQLS_LockHandler
      Get
         Throw New NotImplementedException("LockHandler-Property has not been provided by the inheritor.")
         Return _cLH
      End Get
      Set(ByVal value As cSQLS_LockHandler)
         _cLH = value
      End Set
   End Property

   ''' <summary>
   ''' Must return an object that may serve as the DataSource for the record-grid
   ''' </summary>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable Function RecordList() As DataTable
      Throw New NotImplementedException("RecordList-Function has not been provided by the inheritor.")
      Return Nothing
   End Function

   ''' <summary>
   ''' Resets all data-controls and either empties their contents or sets the default values.
   ''' </summary>
   ''' <remarks></remarks>
   Friend Overridable Sub EmptyControls()
      Throw New NotImplementedException("EmptyControls-Method has not been provided by the inheritor.")
   End Sub

   ''' <summary>
   ''' Sets up the form's caption.
   ''' </summary>
   ''' <param name="strAppendText">Allows for specifying that the form is i.e. in edit-state.
   ''' Leaving this argument or passing an empty string will reset to the regular caption.</param>
   ''' <remarks></remarks>
   Friend Overridable Sub FormCaption(Optional ByVal strAppendText As String = "")
      Throw New NotImplementedException("FormCaption-Method has not been provided by the inheritor.")
      Me.Text = "Not set!"
   End Sub

   Friend Overridable Function IsDataValid() As Boolean
      Throw New NotImplementedException("IsDataValid-Function has not been provided by the inheritor.")
      Return False
   End Function

   ''' <summary>
   ''' This function is to actually save any pending changes. The function MUST return whether or not data has been saved successfully or not!
   ''' </summary>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Friend Overridable Function FormData_Save(Optional ByRef fShowMessage As Boolean = True) As Boolean
      Throw New NotImplementedException("FormData_Save-Function has not been provided by the inheritor.")
      Return False
   End Function

   ''' <summary>
   ''' This function is to actually load data into the controls. 
   ''' </summary>
   ''' <remarks></remarks>
   Friend Overridable Sub FormData_LoadRecord(ByVal Record_ID As Integer)
      Throw New NotImplementedException("FormData_LoadRecord-Method has not been provided by the inheritor.")
   End Sub

   Friend Overridable Function FormData_EditModeEntering_LockCreated() As Boolean
      Throw New NotImplementedException("FormData_EditModeEntering_LockCreated-Method has not been provided by the inheritor.")
      Return False
   End Function

   Friend Overridable Sub FormData_EditModeCanceling_RemoveLock()
      Throw New NotImplementedException("FormData_EditModeCanceling-Method has not been provided by the inheritor.")
   End Sub

   ''' <summary>
   ''' 
   ''' </summary>
   ''' <param name="fLockControls"></param>
   ''' <remarks></remarks>
   Friend Overridable Sub DataControls_SetLockstate(ByVal fLockControls As Boolean)
      Throw New NotImplementedException("DataControls_SetLockstate-Method has not been provided by the inheritor.")
   End Sub

#End Region

#Region " --- Events raised by this class --- "

   ''' <summary>
   ''' Raise, when the local LockHandler class automatically removed a record-lock (i.e. due to a timeout)
   ''' </summary>
   Friend Event EditState_LockAutoRemoved(ByVal TableName As String, ByVal RecordID As Integer, ByVal LockAutoRemoved As Boolean)

   ''' <summary>
   ''' Raised, when the user is about to enter the edit-mode, but before editing takes place.
   ''' </summary>
   ''' <param name="Cancel">If set to True, the locking of a record will be prevented, hence editing will be canceled.</param>
   ''' <remarks></remarks>
   Friend Event EditState_BeforeEnter(ByRef Cancel As Boolean)

   ''' <summary>
   ''' Raised, after the has entered the edit-mode (i.e., after the record-lock has been created).
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event EditState_AfterEnter()

   ''' <summary>
   ''' Raised, before the current record is about to be deleted
   ''' </summary>
   ''' <param name="Record_ID__ToBeDeleted"></param>
   ''' <param name="Cancel">If set to False by caller, the deletion will be aborted.</param>
   ''' <remarks></remarks>
   Friend Event Record_BeforeDelete(ByVal Record_ID__ToBeDeleted As Integer, ByRef Cancel As Boolean)

   ''' <summary>
   ''' Raised, when the current record is to be deleted
   ''' </summary>
   ''' <param name="Record_ID__ToBeDeleted"></param>
   ''' <remarks></remarks>
   Friend Event Record_Delete(ByVal Record_ID__ToBeDeleted As Integer)

   ''' <summary>
   ''' Fired when the Form's Load-event is starting, but before any code here is being executed.
   ''' If there is any controls such as i.e. a combobox with items bound to data that is being edited in the form,
   ''' create and assign any those controls' DataSources when this event is being fired.
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event FormLoading()

   ''' <summary>
   ''' Fired when the local Load-event is about to end (last statement in BaseForm_Load)
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event FormReady()

   ''' <summary>
   ''' Will be fired before the form's dirty-status changes
   ''' </summary>
   ''' <param name="NewState"></param>
   ''' <remarks></remarks>
   Friend Event DirtyStateChanged(ByVal NewState As Boolean)

   ''' <summary>
   ''' Will be fired after the form's dirty-status has changed
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event DirtyState_BeforeChange()

   ''' <summary>
   ''' Will be fired after the value of the NewRecord-property has changed
   ''' </summary>
   ''' <param name="NewState"></param>
   ''' <remarks></remarks>
   Friend Event NewRecord_StateChanged(ByVal NewState As Boolean)

   ''' <summary>
   ''' Fired after a record has been saved successfully
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event RecordSaved()

   ''' <summary>
   ''' Fired after a record has been selected in the Grid Control
   ''' </summary>
   ''' <remarks></remarks>
   Friend Event RecordSelected(ByVal sender As Object, ByVal e As System.EventArgs)

#End Region

#Region " --- Member Variables --- "

   'Lokaler Lock-Handler
   Friend WithEvents _cLH As cSQLS_LockHandler = Nothing

   Private WithEvents _GridControl As DataGridView

   'Common flags, most of them published via properties
   Private _fDirty_Form As Boolean = Nothing
   Private _fNewRecord As Boolean = False
   Private _fGridView_Setting_DataSource As Boolean = False
   Private _fEditStateActive As Boolean = False

   'Pointer to the current record (published via the Record_ID-property)
   Private _Record_ID As Integer = 0

   'Will contain the record that was selected last (published via the RowIndex_LastSelected-property)
   'Since, if a new record is being edited, the selection in the grid is cleared, this variable will
   'help to set the grid to the last shown record in case the user cancels editing/adding.
   Private _intRowIndex_LastSelected As Integer = -1

   'Until at least one record is being saved (in which case the DialogResult will be set to <OK>) OR a record has been deleted, 
   'the form's result is to show <Cancel>.
   Private _DialogResult As DialogResult = Windows.Forms.DialogResult.Cancel

   Private _strCreateUser As String = String.Empty
   Private _strUpdateUser As String = String.Empty

#End Region

#Region " --- Properties --- "

   ''' <summary>
   ''' The user that created a record.
   ''' </summary>
   Public Property CreateUser() As String
      Get
         Return _strCreateUser
      End Get
      Set(ByVal value As String)
         _strCreateUser = value
      End Set
   End Property

   ''' <summary>
   ''' The user that last saved a record.
   ''' </summary>
   ''' <value></value>
   Public Property UpdateUser() As String
      Get
         Return _strUpdateUser
      End Get
      Set(ByVal value As String)
         _strUpdateUser = value
      End Set
   End Property

   ''' <summary>
   ''' Returns the value of cell with the specified column-index, in the first adjacent row (mode: previous first, then next row).
   ''' </summary>
   ''' <param name="dg">Reference to the grid-control</param>
   ''' <param name="intColIndex">Index of the column whose value is to be returned</param>
   ''' <value></value>
   ''' <returns>The value of the requested column, if there is other rows in the grid control. If no other rows are found, Object.Nothing will be returned.</returns>
   ''' <remarks></remarks>
   Public ReadOnly Property Grid_ColumnValue_SelectedRecord_PreviousOrNext(ByVal dg As DataGridView, ByVal intColIndex As Integer) As Object
      Get
         'The PK of the record that is to be returned, based on the one that is currently being selected
         Dim Record_ID_ToBeReturned As String = String.Empty
         'Retrieve the index of the row that appears before the item that is to be deleted
         Dim intRowIndex_ShowRecord As Integer = dg.SelectedRows(0).Index - 1
         If intRowIndex_ShowRecord >= 0 Then
            Return dg.Rows(intRowIndex_ShowRecord).Cells(intColIndex).Value
         Else
            'The current item is the first one in the list, so attempt to select the one after the currently selected one
            If dg.Rows.Count - 1 > dg.SelectedRows(0).Index Then
               intRowIndex_ShowRecord = dg.SelectedRows(0).Index + 1
               Return dg.Rows(intRowIndex_ShowRecord).Cells(intColIndex).Value
            Else
               'Cannot find another record
               Return Nothing
            End If
         End If
      End Get
   End Property

   ''' <summary>
   ''' Sets/returns the current edit-state.
   ''' </summary>
   ''' <value></value>
   ''' <returns>True: The current record is being edited</returns>
   ''' <remarks></remarks>
   Public Property EditStateActive() As Boolean
      Get
         Return _fEditStateActive
      End Get
      Private Set(ByVal value As Boolean)
         _fEditStateActive = value
      End Set
   End Property

   ''' <summary>
   ''' Dirty-Tracker for form-data
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Overridable Property Dirty() As Boolean
      Get
         Return _fDirty_Form
      End Get
      Set(ByVal value As Boolean)
         'If value = True Then Stop

         'Prevent unnecessary actions
         Dim fValueChanged As Boolean = (_fDirty_Form <> value)
         If Not fValueChanged Then
            'For debugging purposes only.
            'Stop
         End If
         If fValueChanged Then RaiseEvent DirtyState_BeforeChange()

         _fDirty_Form = value
         If Not _fDirty_Form Then
            Controls_SetLockState(fLockControls:=True)
            NewRecord = False
         Else
            RecordCommands_DisEnable()
         End If
         'Controls_SetLockState(fLockControls:=(Not _fDirty_Form))
         'RecordCommands_DisEnable()
         If fValueChanged Then RaiseEvent DirtyStateChanged(_fDirty_Form)
      End Set
   End Property

   ''' <summary>
   ''' NewRecord-tracker for form-data
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks></remarks>
   Overridable Property NewRecord() As Boolean
      Get
         Return _fNewRecord
      End Get
      Set(ByVal value As Boolean)
         _fNewRecord = value
         'Prevent unnecessary actions
         'If _fNewRecord <> value Then RaiseEvent NewRecord_StateChanged(_fNewRecord)
         RaiseEvent NewRecord_StateChanged(_fNewRecord)
      End Set
   End Property

   ''' <summary>
   ''' Sets or returns the current record's PK
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks>This property may not be overrided.</remarks>
   Friend Property Record_ID() As Integer
      Get
         Return _Record_ID
      End Get
      Set(ByVal value As Integer)
         'Prevent unnecessary actions
         'If _Record_ID = value Then ...

         'Prevent the value from being changed while the grid's DataSource-property is being set
         'If _fGridView_Setting_DataSource AndAlso Not _Record_ID.Equals(String.Empty) Then Return

         _Record_ID = value
      End Set
   End Property

   ''' <summary>
   '''Sets/returns the row-index of the record that was selected last in the user-grid 
   ''' </summary>
   ''' <value></value>
   ''' <returns></returns>
   ''' <remarks>This property may not be overrided.</remarks>
   Friend Property RowIndex_LastSelected() As Integer
      Get
         Return _intRowIndex_LastSelected
      End Get
      Set(ByVal value As Integer)
         _intRowIndex_LastSelected = value
      End Set
   End Property

#End Region

#Region " --- Helpers --- "

   Overridable Sub ReloadGrid()
      Me.Cursor = Cursors.WaitCursor

      If GridControl Is Nothing Then Return

      'Setup the grid's DataSource, assuring that the current resp. last active (local) Record-ID/String is not being changed
      _fGridView_Setting_DataSource = True
      GridControl.ClearSelection()
      GridControl.DataSource = RecordList()
      _fGridView_Setting_DataSource = False

      If Not Record_ID.Equals(String.Empty) Then
         'Attempt to select the record that shows the current form's PK
         For intRow As Integer = 0 To GridControl.Rows.Count - 1
            If CType(GridControl.Rows(intRow).Cells(Grid_ColumnIndex_Record_ID).Value, Integer) = Record_ID Then
               GridControl.Rows(intRow).Selected = True
               GridControl_SelectionChanged(Nothing, Nothing)
               Exit For
            End If
         Next
      End If

      If GridControl.Rows.Count > 0 AndAlso GridControl.SelectedRows.Count < 1 Then
         'No Record-PK set so select the very first row
         GridControl.Rows(0).Selected = True
      Else
         'No record to be selected, so assure that all controls are emtpy
         EmptyControls()
      End If
      GridControl_SelectionChanged(Nothing, Nothing)

      'Clear the dirty flag (this will lock all data-controls for editing as well)
      Me.Dirty = False
      Me.Cursor = Cursors.Default
   End Sub

   Friend Function Check_Save() As Boolean
      If Me.Dirty Then
         Select Case MsgBox("You have changed the current " & RecordDesignator_Singular & "." & vbCrLf & vbCrLf & _
               "Would you like to save your changes?", MsgBoxStyle.Question Or MsgBoxStyle.YesNoCancel Or MsgBoxStyle.DefaultButton3, Application.ProductName)
            Case MsgBoxResult.Yes
               'Save pending changes and continue
               cmdSave_Click(Nothing, Nothing)
               'Return True only if data was saved successfully
               Return (Not Me.Dirty)
            Case MsgBoxResult.No
               'Go on without saving the changes 
               cmdCancelEdit_Click(Nothing, Nothing)
               Return True
            Case Else
               'Back to editing the record
               Return False
         End Select
      Else
         'No pending changes
         Return True
      End If
   End Function

   ''' <summary>
   ''' Allows or prevents editing of controls in this form that are bound to data.
   ''' </summary>
   ''' <param name="fLockControls">True: prevent editing; False: allow editing.</param>
   ''' <remarks></remarks>
   Overridable Sub Controls_SetLockState(ByVal fLockControls As Boolean)
      'Set the focus according to the requested lock-state
      If fLockControls Then
         GridControl.Focus()
      Else
         'Set the focus to the first data-control
         'Me.SelectNextControl(GridControl, True, True, False, True)
      End If

      'Dis-/enable editing in data-controls
      DataControls_SetLockstate(fLockControls)
      GridControl.Enabled = (fLockControls And GridControl.RowCount > 0)

      'Show/hide data-buttons according to the current lock-state
      cmdSave.Visible = (Not fLockControls)
      cmdCancelEdit.Visible = (Not fLockControls)
      cmdNew.Visible = fLockControls
      cmdDelete.Visible = fLockControls
      cmdEdit.Visible = fLockControls

      RecordCommands_DisEnable()
   End Sub

   ''' <summary>
   ''' Will, in addition to showing or hiding controls as done in Controls_SetLockState, dis- or enable controls
   ''' according to whether a record is currently being edited and/or a record being selected.
   ''' </summary>
   ''' <remarks></remarks>
   Overridable Sub RecordCommands_DisEnable()
      Dim fFormIsDirty As Boolean = Me.Dirty
      Dim fRecordIsSelected As Boolean = (GridControl.SelectedRows.Count > 0)

      cmdDelete.Enabled = fRecordIsSelected
      cmdEdit.Enabled = fRecordIsSelected
      'Alternative:
      'cmdSave.Enabled = (fRecordIsSelected AndAlso fFormIsDirty)
      cmdSave.Enabled = fFormIsDirty
      cmdNew.Enabled = (Not fFormIsDirty)
   End Sub

#End Region

#Region " --- Event Handlers --- "

   Private Sub LockHandler_LockRemoved(ByVal TableName As String, ByVal RecordID As Integer, ByVal LockAutoRemoved As Boolean) Handles _cLH.LockRemoved
      RaiseEvent EditState_LockAutoRemoved(TableName, Record_ID, LockAutoRemoved)
   End Sub

   Overridable Sub BaseForm_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
      'Data in this form has been changed ...
      If e.CloseReason = CloseReason.UserClosing Then
         'The user is closing the form, so ask whether to save any pending changes
         If Not Check_Save() Then e.Cancel = True
      Else
         'For all other reasons (ApplicationExitCall, FormOwnerClosing, MdiFormClosing, None, TaskManagerClosing, WindowsShutDown),
         'simply cancel editing and continue.
         cmdCancelEdit_Click(sender, Nothing)
      End If
      'End If
      If Me.EditStateActive Then cmdCancelEdit_Click(sender, Nothing)
      If Not e.Cancel AndAlso Me.Modal Then Me.DialogResult = _DialogResult
   End Sub

   Overridable Sub BaseForm_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
      'These properties are not being used in this sample.
      'ToolTip.Active = My.Settings.ShowTooltips
      'HelpProvider.HelpNamespace = cHelpAccess.HelpFile_FullFilename

      'Since this form is to be inherited, the Windows Forms Designer cannot retrieve a reference to the GridControl at
      'design-time, so exit right away, if the form is being designed.
      If Me.DesignMode Then Return

      'Pull the reference to the grid-control
      _GridControl = GridControl

      RaiseEvent FormLoading()

      _GridControl.AutoGenerateColumns = False

      tsslCreator.Text = String.Empty
      tsslUpdater.Text = String.Empty
      EmptyControls()
      ToolTip.ToolTipTitle = RecordDesignator_Plural & " Administration"

      ReloadGrid()

      FormCaption()

      RaiseEvent FormReady()
   End Sub

   Overridable Sub GridControl_EnabledChanged(ByVal sender As Object, ByVal e As System.EventArgs) _
            Handles _GridControl.EnabledChanged
      cGeneral.GridView_EnabledChanged(sender)
   End Sub

   Overridable Sub GridControl_RowValidating(ByVal sender As Object, ByVal e As System.Windows.Forms.DataGridViewCellCancelEventArgs) _
            Handles _GridControl.RowValidating
      If Not Check_Save() Then
         e.Cancel = True
      End If
   End Sub

   Overridable Sub GridControl_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) _
            Handles _GridControl.SelectionChanged
      With GridControl
         If _fGridView_Setting_DataSource Then Return

         If .SelectedRows.Count > 0 Then
            'Different user selected - load this record into the controls
            Record_ID = CType(.SelectedRows(0).Cells(Grid_ColumnIndex_Record_ID).Value, Integer)
            FormData_LoadRecord(Record_ID)
            Me.Dirty = False
            RowIndex_LastSelected = .SelectedRows(0).Index
            'Assure the row is visible (i.e. scroll it into view, if required)
            If (.Rows(RowIndex_LastSelected).State And DataGridViewElementStates.Displayed) <> DataGridViewElementStates.Displayed Then ' .Rows(RowIndex_LastSelected).State <> DataGridViewElementStates.Displayed Then
               Dim intColIndex_FirstVisible As Integer = 0
               For intCounter As Integer = 0 To .Columns.Count - 1
                  If .Columns(intCounter).Visible AndAlso .Columns(intCounter).Width > 0 Then
                     intColIndex_FirstVisible = intCounter
                     Exit For
                  End If
               Next
               .FirstDisplayedCell = .Rows(RowIndex_LastSelected).Cells(intColIndex_FirstVisible)
            End If
         End If
      End With
      RaiseEvent RecordSelected(sender, e)
   End Sub

   ''' <summary>
   ''' This method must be added to all data-controls' [whatever_]changed-handler in order for the form's Dirty-property to be appropriate.
   ''' </summary>
   ''' <param name="sender"></param>
   ''' <param name="e"></param>
   ''' <remarks></remarks>
   Friend Sub Form_Dirty(ByVal sender As System.Object, ByVal e As System.EventArgs)
      Me.Dirty = True
   End Sub

   Friend Sub cmdSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdSave.Click
      Dim fSuccess As Boolean = True

      'A. Save user-data (if it has been changed)
      If Me.Dirty Then
         'Data valid?
         If Not IsDataValid() Then Return

         Me.Cursor = Cursors.WaitCursor
         fSuccess = FormData_Save()
      End If
      If Not fSuccess Then
         MsgBox("The current " & RecordDesignator_Singular & "-record could not be saved!", MsgBoxStyle.Exclamation, Application.ProductName & ": Save failed!")
         Me.Cursor = Cursors.Default
         Return
      Else
         'Any other / additional stuff could go here (i.e. raise an event to the heir)
         Me.Dirty = False
         EditStateActive = False
         RaiseEvent RecordSaved()
         'The form's result is to show OK after any record has been changed
         _DialogResult = Windows.Forms.DialogResult.OK
      End If
      ReloadGrid()
      Me.Cursor = Cursors.Default
   End Sub

   Friend Sub cmdNew_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdNew.Click
      'Empty all controls, create a PK for the new record, unlock the form and focus the first control
      tsslCreator.Text = String.Empty
      tsslUpdater.Text = String.Empty
      EmptyControls()
      GridControl.ClearSelection()
      Me.Dirty = False
      Record_ID = 0
      NewRecord = True
      Controls_SetLockState(fLockControls:=False)
      EditStateActive = True
      FormCaption("New record")
   End Sub

   Friend Sub cmdEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdEdit.Click
      Dim fCancelEditing As Boolean = False
      RaiseEvent EditState_BeforeEnter(fCancelEditing)
      'Cancel if the receiver of the event flagged accordingly
      If fCancelEditing Then Return

      'Attempt to lock the record for editing. If possible, unlock all controls
      Dim fRecordLocked As Boolean = FormData_EditModeEntering_LockCreated() ' _cData_User.Edit_Record(Record_ID)
      Controls_SetLockState(fLockControls:=Not fRecordLocked)
      If fRecordLocked Then FormCaption("Edit record")

      Me.EditStateActive = True
      RaiseEvent EditState_AfterEnter()
   End Sub

   Friend Sub cmdCancelEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdCancelEdit.Click
      'Remove the lock and reload the form
      If Not Me.NewRecord Then FormData_EditModeCanceling_RemoveLock() ' _cData_User.Cancel_Edit(Record_ID)
      Controls_SetLockState(fLockControls:=True)
      If GridControl.SelectedRows.Count < 1 Then
         If RowIndex_LastSelected > 0 Then
            'Select the previously selected record
            GridControl.Rows(RowIndex_LastSelected).Selected = True
         ElseIf GridControl.Rows.Count > 0 Then
            'Select the first user
            GridControl.Rows(0).Selected = True
         End If
      End If
      GridControl_SelectionChanged(sender, e)
      Me.Dirty = False
      EditStateActive = False
      FormCaption()
   End Sub

   Private Sub cmdDelete_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmdDelete.Click
      Dim fCancelDeletion As Boolean = False
      RaiseEvent Record_BeforeDelete(Record_ID, fCancelDeletion)
      If fCancelDeletion Then Return

      If MsgBox("You are about to delete a " & RecordDesignator_Singular & "-record from the database." & vbCrLf & _
         "You will not be able to undo this operation." & vbCrLf & vbCrLf & _
         "Are sure you want to delete this " & RecordDesignator_Singular & "?", _
         MsgBoxStyle.Question Or MsgBoxStyle.YesNo Or MsgBoxStyle.DefaultButton2, Application.ProductName & ": Delete record") _
         = MsgBoxResult.No Then Return

      RaiseEvent Record_Delete(Record_ID)

      'The form's result is to show OK after any record has been changed, this includes the deletion of any records
      _DialogResult = Windows.Forms.DialogResult.OK
   End Sub

   Private Sub cmdClose_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmdClose.Click
      'The Form_Closing event will take care of actions due to any pending changes.
      If Not Check_Save() Then Return
      Me.Close()
   End Sub

#End Region

End Class